#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "symbol.h"
#include "value.h"
#include "asm.h"
#include "asm_syntax.h"
#include "asm_symbol.h"

static int reg_next = 0;

static int reg_reserve(void)
{
  int r;
  r = reg_next++;
  return r;
}

static int reg_release(void)
{
  --reg_next;
  return reg_next;
}

static int reg_current(void)
{
  return reg_next - 1;
}

static char *reg_name(int reg)
{
  static char *name[] = {
    "Register Error",
    "%ebx",
    "32(%esp)", "36(%esp)", "40(%esp)", "44(%esp)",
    "48(%esp)", "52(%esp)", "56(%esp)", "60(%esp)",
  };
  return name[reg + 1];
}

static char *reg_name_current(void)
{
  return reg_name(reg_current());
}

static int label_next = 1;

static int label_reserve(int n)
{
  int r;
  r = label_next;
  label_next += n;
  return r;
}

static int label_release(int n)
{
  label_next -= n;
  return label_next;
}

static int label_current(void)
{
  return label_next - 1;
}

static int reg_stack_size(void)
{
  return sizeof(int) * 16;
}

static char *r0_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "al";
  case  2: return "ax";
  default: break;
  }
  return "eax";
}

static char *r1_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "dl";
  case  2: return "dx";
  default: break;
  }
  return "edx";
}

static char *r2_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "cl";
  case  2: return "cx";
  default: break;
  }
  return "ecx";
}

static char *mov_by_size(model_t model)
{
  return "mov";
}

static int sign_extension(model_t model)
{
  if (!model_is_address(model)) {
    switch (model->type) {
    case C_TYPE_CHAR:
      if (model->flags & MODEL_FLAG_UNSIGNED)
	printf("\tmovzbl\t%%al,\t%%eax\n");
      else
	printf("\tmovsbl\t%%al,\t%%eax\n");
      break;
    case C_TYPE_SHORT:
      if (model->flags & MODEL_FLAG_UNSIGNED)
	printf("\tmovzwl\t%%ax,\t%%eax\n");
      else
	printf("\tmovswl\t%%ax,\t%%eax\n");
      break;
    default:
      break;
    }
  }

  return 0;
}

int asm_symbol_address_get(symbol_t symbol)
{
  asm_section_change_text();

  printf("\t/* ADDRESS GET BEGIN (%s, %s) */\n", symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    /* fall through */
  case C_TYPE_VARIABLE:
    switch (symbol->location.type) {
    case C_TYPE_STATIC:
      printf("\tmov\t$%s, %%eax\n", symbol->name);
      break;
    case C_TYPE_STACK:
      printf("\tlea\t%d(%%esp), %%eax\n",
	     reg_stack_size() + symbol->location.offset);
      break;
    case C_TYPE_STRUCT:
    case C_TYPE_UNION:
      ERREXIT(1);
    case C_TYPE_SYMBOL:
      asm_symbol_address_get(symbol->location.base.symbol);
      printf("\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_POINTER:
      asm_symbol_value_get(symbol->location.base.symbol);
      printf("\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_NONE:
      printf("\tmov\t$%s, %%eax\n", symbol->name);
      break;
    default:
      ERREXIT(1);
    }
    break;
  default:
    ERREXIT(1);
  }

  printf("\t/* ADDRESS GET END (%s) */\n", symbol->name);
  return 0;
}

static int asm_symbol_value_op(symbol_t symbol)
{
  symbol_t *arg;
  model_t model;
  int i, n;

  printf("\t/* VALUE OP BEGIN (%s, %s, %s) */\n",
	 symbol->name, type_get_word(symbol->type),
	 type_get_word(symbol->value->type));

  switch (symbol->value->type) {
  case C_TYPE_NUMBER:
    printf("\tmov\t$0x%x, %%eax\n", symbol->value->obj.number);
    goto ret;
  default:
    break;
  }

  arg = symbol->value->obj.op.arg;

  switch (symbol->value->type) {
  case C_TYPE_CAST:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    break;
  case C_TYPE_ADDRESS:
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    break;
  case C_TYPE_POINTER:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    model = model_copy(arg[0]->model);
    if (model_del_modifier(model) < 0)
      ERREXIT(1);
    printf("\t%s\t(%%eax), %%%s\n", mov_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_DOT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    printf("\t%s\t%d(%%eax), %%%s\n",
	   mov_by_size(symbol->model), symbol->location.offset, r0_by_size(symbol->model));
    break;
  case C_TYPE_NOT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    label_reserve(2);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    printf("\tje\t%df\n", label_current() - 1);
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_INV:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tnot\t%%eax\n");
    break;
  case C_TYPE_PLUS:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    break;
  case C_TYPE_MINUS:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tneg\t%%eax\n");
    break;
  case C_TYPE_INC: /* ++x */
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    model = arg[0]->model;
    printf("\tmov\t%%eax, %%edx\n");
    printf("\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    printf("\tadd\t$%d, %%%s\n", n, r0_by_size(model));
    printf("\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_DEC: /* --x */
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    model = arg[0]->model;
    printf("\tmov\t%%eax, %%edx\n");
    printf("\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    printf("\tsub\t$%d, %%%s\n", n, r0_by_size(model));
    printf("\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_INCA: /* x++ */
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    model = arg[0]->model;
    printf("\tmov\t%%eax, %%edx\n");
    printf("\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    printf("\tadd\t$%d, %%%s\n", n, r0_by_size(model));
    printf("\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    printf("\tmov\t%s, %%eax\n", reg_name_current());
    reg_release();
    break;
  case C_TYPE_DECA: /* x-- */
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    model = arg[0]->model;
    printf("\tmov\t%%eax, %%edx\n");
    printf("\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    printf("\tsub\t$%d, %%%s\n", n, r0_by_size(model));
    printf("\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    printf("\tmov\t%s, %%eax\n", reg_name_current());
    reg_release();
    break;

  case C_TYPE_ADD:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\tadd\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_SUB:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\tsub\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_MUL:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\timul\t%%%s\n", r1_by_size(model));
    break;
  case C_TYPE_DIV:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tmov\t$0, %%edx\n");
    } else {
      printf("\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tdiv\t%%%s\n", r2_by_size(arg[1]->model));
    } else {
      printf("\tidiv\t%%ecx\n");
    }
    break;
  case C_TYPE_MOD:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tmov\t$0, %%edx\n");
    } else {
      printf("\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tdiv\t%%ecx\n");
    } else {
      printf("\tidiv\t%%ecx\n");
    }
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_AND:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\tand\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_OR:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\tor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_XOR:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\txor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_LSHIFT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    printf("\tshl\t%%cl, %%eax\n");
    break;
  case C_TYPE_RSHIFT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%cl, %%eax\n",
	   (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) ? "shr" : "sar");
    break;

  case C_TYPE_LT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tjb\t%df\n", label_current() - 1);
    } else {
      printf("\tjl\t%df\n", label_current() - 1);
    }
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_GT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tja\t%df\n", label_current() - 1);
    } else {
      printf("\tjg\t%df\n", label_current() - 1);
    }
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_LTEQ:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tjbe\t%df\n", label_current() - 1);
    } else {
      printf("\tjle\t%df\n", label_current() - 1);
    }
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_GTEQ:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tjae\t%df\n", label_current() - 1);
    } else {
      printf("\tjge\t%df\n", label_current() - 1);
    }
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_EQEQ:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tje\t%df\n", label_current() - 1);
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_NEQ:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    printf("\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tjne\t%df\n", label_current() - 1);
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;

  case C_TYPE_EQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_ADDEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\tadd\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_SUBEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\tsub\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_MULEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\timul\t%%%s\n", r1_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_DIVEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tmov\t$0, %%edx\n");
    } else {
      printf("\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tdiv\t%%%s\n", r2_by_size(arg[1]->model));
    } else {
      printf("\tidiv\t%%ecx\n");
    }
    sign_extension(symbol->model);
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_MODEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tmov\t$0, %%edx\n");
    } else {
      printf("\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      printf("\tdiv\t%%ecx\n");
    } else {
      printf("\tidiv\t%%ecx\n");
    }
    printf("\tmov\t%%edx, %%eax\n");
    sign_extension(symbol->model);
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_LSHIFTEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    printf("\tshl\t%%cl, %%eax\n");
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_RSHIFTEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name_current());
    printf("\t%s\t%%cl, %%eax\n",
	   (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) ? "shr" : "sar");
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_ANDEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\tand\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_OREQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\tor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_XOREQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    model = arg[0]->model;
    reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    printf("\txor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    printf("\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    printf("\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    printf("\tmov\t%%edx, %%eax\n");
    break;

  case C_TYPE_ANDAND:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    label_reserve(2);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    printf("\tje\t%df\n", label_current() - 1);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[1]->model));
    printf("\tje\t%df\n", label_current() - 1);
    printf("\tmov\t$1, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$0, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_OROR:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    label_reserve(2);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    printf("\tjne\t%df\n", label_current() - 1);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[1]->model));
    printf("\tjne\t%df\n", label_current() - 1);
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\tmov\t$1, %%eax\n", label_current() - 1);
    printf("%d:\n", label_current());
    label_release(2);
    break;

  case C_TYPE_QUESTION:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    label_reserve(2);
    printf("\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    printf("\tje\t%df\n", label_current() - 1);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[1]);
    printf("\tjmp\t%df\n", label_current());
    printf("%d:\n", label_current() - 1);
    asm_symbol_value_set(arg[2]);
    asm_symbol_value_get(arg[2]);
    printf("%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_COMMA:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_get(arg[0]);
    asm_symbol_value_get(arg[1]);
    break;

  case C_TYPE_ARG:
    asm_symbol_value_get(symbol);
    break;
  case C_TYPE_FUNCTION:
    for (i = 0; i < symbol->value->obj.function.args.number; i++) {
      asm_symbol_value_set(symbol->value->obj.function.args.arg[i]);
    }
    for (i = 0; i < symbol->value->obj.function.args.number; i++) {
      asm_symbol_value_get(symbol->value->obj.function.args.arg[i]);
      reg_reserve();
      printf("\tmov\t%%eax, %s\n", reg_name_current());
    }
    for (i = 0; i < symbol->value->obj.function.args.number; i++) {
      printf("\tmov\t%s, %%eax\n", reg_name_current());
      printf("\tmov\t%%eax, 0x%x(%%esp)\n",
	     (symbol->value->obj.function.args.number - 1 - i) * 4);
      reg_release();
    }

    if (model_is_pointer(symbol->value->obj.function.symbol->model)) {
      asm_symbol_address_get(symbol->value->obj.function.symbol);
      printf("\tmov\t(%%eax), %%eax\n");
      printf("\tcall\t*%%eax\n");
    } else {
      printf("\tcall\t%s\n", symbol->value->obj.function.symbol->name);
    }
    break;
  default:
    ERREXIT(1);
  }

ret:
  sign_extension(symbol->model);
  printf("\t/* VALUE OP END (%s, %s, %s) */\n",
	 symbol->name, type_get_word(symbol->type),
	 type_get_word(symbol->value->type));
  return 0;
}

int asm_symbol_value_set(symbol_t symbol)
{
  asm_section_change_text();

  printf("\t/* VALUE SET BEGIN (%s, %s) */\n",
	 symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    break;
  case C_TYPE_VARIABLE:
    if (symbol->flags & SYMBOL_FLAG_TEMPORARY) {
      asm_symbol_value_op(symbol);
      if (symbol->flags & SYMBOL_FLAG_LOCATED) {
	reg_reserve();
	printf("\tmov\t%%eax, %s\n", reg_name_current());
	asm_symbol_address_get(symbol);
	printf("\tmov\t%s, %%edx\n", reg_name_current());
	reg_release();
	printf("\t%s\t%%%s, (%%eax)\n",
	       mov_by_size(symbol->model), r1_by_size(symbol->model));
      }
    }
    break;
  default:
    ERREXIT(1);
  }

  printf("\t/* VALUE SET END (%s) */\n", symbol->name);
  return 0;
}

int asm_symbol_value_get(symbol_t symbol)
{
  asm_section_change_text();

  printf("\t/* VALUE GET BEGIN (%s, %s) */\n",
	 symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    asm_symbol_address_get(symbol);
    break;
  case C_TYPE_VARIABLE:
    if (symbol->flags & SYMBOL_FLAG_READONLY) {
      switch (symbol->value->type) {
      case C_TYPE_NUMBER:
	printf("\t%s\t$%d, %%%s\n", mov_by_size(symbol->model),
	       symbol->value->obj.number, r0_by_size(symbol->model));
	break;
      default:
	ERREXIT(1);
      }
      break;
    }
    if (symbol->flags & SYMBOL_FLAG_LOCATED) {
      asm_symbol_address_get(symbol);
      printf("\t%s\t(%%eax), %%%s\n",
	     mov_by_size(symbol->model), r0_by_size(symbol->model));
      break;
    }
    ERREXIT(1);
  default:
    ERREXIT(1);
  }

  sign_extension(symbol->model);
  printf("\t/* VALUE GET END (%s) */\n", symbol->name);
  return 0;
}

static int asm_symbol_label_value(symbol_t symbol);
static int asm_symbol_variable_value(symbol_t symbol);

static int asm_symbol_array(objlist_t symbols)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(symbols);
       !objlist_is_end(symbols, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
#if 1
    printf("\t.align\t%d\n", model_get_align(symbol->model));
    switch (symbol->type) {
    case C_TYPE_LABEL:    asm_symbol_label_value(symbol);    break;
    case C_TYPE_VARIABLE: asm_symbol_variable_value(symbol); break;
    default:
      ERREXIT(1);
    }
#else
    asm_symbol(symbol);
#endif
  }

  return 0;
}

static int asm_symbol_area(area_t area)
{
  printf("\t.align\t%d\n", area->align);
  printf("\t.space\t%d, 0\n", area->size);
  return 0;
}

static int asm_symbol_function(symbol_t symbol)
{
  syntax_t syntax;
  objentry_t entry;
  symbol_t s;
  int n;

  asm_section_change_data();

  printf("%s:\n", symbol->obj.function.initlabel->obj.label.name);
  asm_symbol_array(symbol->obj.function.stack->symbols);
  printf("\n");

  asm_section_change_text();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.type\t%s, @function\n", symbol->name);
  printf("%s:\n", symbol->name);

  if ((symbol->value != NULL) &&
      (symbol->value->type == C_TYPE_BLOCK)) {
    syntax = symbol->value->obj.block.syntax;
    if (syntax->type == C_TYPE_BLOCK) {
      printf("\tpush\t%%ebp\n");
      printf("\tmov\t%%esp, %%ebp\n");
      printf("\tpush\t%%esi\n");
      printf("\tpush\t%%edi\n");
      printf("\tpush\t%%ebx\n");

      printf("\tsub\t$0x%x, %%esp\n",
	     reg_stack_size() + symbol->obj.function.stack->size);

      printf("\tlea\t0x%x(%%esp), %%ebx\n", reg_stack_size());
      printf("\tmov\t$%s, %%ecx\n", symbol->obj.function.initlabel->obj.label.name);
      printf("\tmov\t$0x%x, %%edx\n", symbol->obj.function.stack->size);
      label_reserve(2);
      printf("%d:", label_current() - 1);
      printf("\tcmp\t$0, %%edx\n");
      printf("\tjle\t%df\n", label_current());
      printf("\tmov\t(%%ecx), %%al\n");
      printf("\tinc\t%%ecx\n");
      printf("\tmov\t%%al, (%%ebx)\n");
      printf("\tinc\t%%ebx\n");
      printf("\tdec\t%%edx\n");
      printf("\tjmp\t%db\n", label_current() - 1);
      printf("%d:", label_current());
      label_release(2);

      n = 0;
      for (entry = objlist_get_head(symbol->obj.function.args);
	   !objlist_is_end(symbol->obj.function.args, entry);
	   entry = objentry_get_next(entry)) {
	s = objentry_get_obj(entry);
	printf("\tmov\t0x%x(%%esp), %%eax\n",
	       reg_stack_size() + symbol->obj.function.stack->size
	       + 4 /* for EBX */
	       + 8 /* for EDI/ESI */
	       + 4 /* for EBP */
	       + 4 /* for return address */
	       + (n++) * 4);
	printf("\tmov\t%%eax, 0x%x(%%esp)\n",
	       reg_stack_size() + s->location.offset);
      }
      asm_syntax_list(syntax->obj.block.syntaxes);
      printf("%s:\n", symbol->obj.function.retlabel->obj.label.name);
      printf("\tadd\t$0x%x, %%esp\n",
	     reg_stack_size() + symbol->obj.function.stack->size);
      printf("\tpop\t%%ebx\n");
      printf("\tpop\t%%edi\n");
      printf("\tpop\t%%esi\n");
      printf("\tleave\n");
      printf("\tret\n\n");
    }
  }

  return 0;
}

static int asm_symbol_label_header(symbol_t symbol)
{
  asm_section_change_data();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.align\t%d\n", model_get_align(symbol->model));
  printf("\t.type\t%s, @object\n", symbol->name);
  printf("\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  printf("%s:\n", symbol->name);

  return 0;
}

static int asm_symbol_string(value_t value)
{
  int i;

  printf("\t.string\t\"");
  for (i = 0; i < value->obj.string.size - 1; i++) {
    switch (value->obj.string.s[i]) {
    case '\0': printf("\\0"); break;
    case '\'': printf("\\\'"); break;
    case '\"': printf("\\\""); break;
    case '\n': printf("\\n"); break;
    case '\t': printf("\\t"); break;
    case '\r': printf("\\r"); break;
    default:   printf("%c", value->obj.string.s[i]); break;
    }
  }
  printf("\"\n");

  return 0;
}

static int asm_symbol_label_value(symbol_t symbol)
{
  if (symbol->value == NULL) {
    printf("\t.space\t%d, 0\t/* %s */\n", model_get_size(symbol->model), symbol->name);
  } else {
    switch (symbol->value->type) {
    case C_TYPE_STRING:
      asm_symbol_string(symbol->value);
      break;
    case C_TYPE_ARRAY:
      asm_symbol_array(symbol->value->obj.array.values);
      printf("\n");
      break;
    default:
      ERREXIT(1);
    }
  }

  return 0;
}

static int asm_symbol_label(symbol_t symbol)
{
  asm_symbol_label_header(symbol);
  asm_symbol_label_value(symbol);
  return 0;
}

static int asm_symbol_variable_header(symbol_t symbol)
{
  asm_section_change_data();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.align\t%d\n", model_get_align(symbol->model));
  printf("\t.type\t%s, @object\n", symbol->name);
  printf("\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  printf("%s:\n", symbol->name);

  return 0;
}

static int asm_symbol_variable_value(symbol_t symbol)
{
  if (model_is_pointer(symbol->model)) {
    printf("\t.long");
  } else {
    switch (symbol->model->type) {
    case C_TYPE_CHAR:  printf("\t.byte" ); break;
    case C_TYPE_SHORT: printf("\t.short"); break;
    case C_TYPE_INT:   printf("\t.int"  ); break;
    case C_TYPE_ENUM:  printf("\t.int"  ); break;
    case C_TYPE_LONG:  printf("\t.long" ); break;
    case C_TYPE_STRUCT:
      if (symbol->value != NULL) {
	printf("\n");
	asm_symbol_array(symbol->value->obj.array.values);
      } else {
	asm_symbol_area(symbol->model->custom._struct->members);
      }
      return 0;
    case C_TYPE_UNION:
      asm_symbol_area(symbol->model->custom._union->members);
      return 0;
    default:
      ERREXIT(1);
    }
  }

  printf("\t");

  if ((symbol->value == NULL) || (symbol->flags & SYMBOL_FLAG_TEMPORARY)) {
    printf("0");
  } else {
    switch (symbol->value->type) {
    case C_TYPE_ARG:    printf("0"); break;
    case C_TYPE_NUMBER: printf("%d", symbol->value->obj.number); break;
    case C_TYPE_LABEL:  printf("%s", symbol->value->obj.label.symbol->name); break;
    default:
      ERREXIT(1);
    }
  }

  if (symbol->name != NULL)
    printf("\t/* %s */", symbol->name);
  printf("\n");

  return 0;
}

static int asm_symbol_variable(symbol_t symbol)
{
  asm_symbol_variable_header(symbol);
  asm_symbol_variable_value(symbol);
  printf("\n");
  return 0;
}

int asm_symbol(symbol_t symbol)
{
  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    if (symbol->flags & SYMBOL_FLAG_FUNCBLOCK) {
      asm_symbol_function(symbol);
      break;
    }
    asm_symbol_label(symbol);
    break;
  case C_TYPE_VARIABLE:
    asm_symbol_variable(symbol);
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}
